import contextlib
import datetime
import logging
from abc import ABCMeta, abstractmethod
from collections import defaultdict
from dataclasses import dataclass, field
from enum import Enum
from functools import partial
from typing import (
    Callable,
    Dict,
    Generic,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Set,
    Type,
    TypeVar,
    Union,
    cast,
)

from pydantic import BaseModel
from typing_extensions import LiteralString

from datahub.configuration.common import ConfigModel
from datahub.configuration.source_common import PlatformInstanceConfigMixin
from datahub.emitter.mcp_builder import mcps_from_mce
from datahub.ingestion.api.closeable import Closeable
from datahub.ingestion.api.common import PipelineContext, RecordEnvelope, WorkUnit
from datahub.ingestion.api.report import Report
from datahub.ingestion.api.source_helpers import (
    auto_browse_path_v2,
    auto_fix_duplicate_schema_field_paths,
    auto_lowercase_urns,
    auto_materialize_referenced_tags_terms,
    auto_status_aspect,
    auto_workunit_reporter,
)
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.metadata.com.linkedin.pegasus2avro.mxe import MetadataChangeEvent
from datahub.metadata.schema_classes import UpstreamLineageClass
from datahub.utilities.lossy_collections import LossyDict, LossyList
from datahub.utilities.type_annotations import get_class_from_annotation

logger = logging.getLogger(__name__)

_MAX_CONTEXT_STRING_LENGTH = 1000


class SourceCapability(Enum):
    PLATFORM_INSTANCE = "Platform Instance"
    DOMAINS = "Domains"
    DATA_PROFILING = "Data Profiling"
    USAGE_STATS = "Dataset Usage"
    PARTITION_SUPPORT = "Partition Support"
    DESCRIPTIONS = "Descriptions"
    LINEAGE_COARSE = "Table-Level Lineage"
    LINEAGE_FINE = "Column-level Lineage"
    OWNERSHIP = "Extract Ownership"
    DELETION_DETECTION = "Detect Deleted Entities"
    TAGS = "Extract Tags"
    SCHEMA_METADATA = "Schema Metadata"
    CONTAINERS = "Asset Containers"
    CLASSIFICATION = "Classification"


class StructuredLogLevel(Enum):
    INFO = logging.INFO
    WARN = logging.WARN
    ERROR = logging.ERROR


@dataclass
class StructuredLogEntry(Report):
    title: Optional[str]
    message: str
    context: LossyList[str]


@dataclass
class StructuredLogs(Report):
    # Underlying Lossy Dicts to Capture Errors, Warnings, and Infos.
    _entries: Dict[StructuredLogLevel, LossyDict[str, StructuredLogEntry]] = field(
        default_factory=lambda: {
            StructuredLogLevel.ERROR: LossyDict(10),
            StructuredLogLevel.WARN: LossyDict(10),
            StructuredLogLevel.INFO: LossyDict(10),
        }
    )

    def report_log(
        self,
        level: StructuredLogLevel,
        message: LiteralString,
        title: Optional[LiteralString] = None,
        context: Optional[str] = None,
        exc: Optional[BaseException] = None,
        log: bool = False,
        stacklevel: int = 1,
    ) -> None:
        """
        Report a user-facing warning for the ingestion run.

        Args:
            level: The level of the log entry.
            message: The main message associated with the report entry. This should be a human-readable message.
            title: The category / heading to present on for this message in the UI.
            context: Additional context (e.g. where, how) for the log entry.
            exc: The exception associated with the event. We'll show the stack trace when in debug mode.
        """

        # One for this method, and one for the containing report_* call.
        stacklevel = stacklevel + 2

        log_key = f"{title}-{message}"
        entries = self._entries[level]

        if context and len(context) > _MAX_CONTEXT_STRING_LENGTH:
            context = f"{context[:_MAX_CONTEXT_STRING_LENGTH]} ..."

        log_content = f"{message} => {context}" if context else message
        if title:
            log_content = f"{title}: {log_content}"
        if exc:
            log_content += f"{log_content}: {exc}"

            if log:
                logger.log(level=level.value, msg=log_content, stacklevel=stacklevel)
                logger.log(
                    level=logging.DEBUG,
                    msg="Full stack trace:",
                    stacklevel=stacklevel,
                    exc_info=exc,
                )

            # Add the simple exception details to the context.
            if context:
                context = f"{context} {type(exc)}: {exc}"
            else:
                context = f"{type(exc)}: {exc}"
        elif log:
            logger.log(level=level.value, msg=log_content, stacklevel=stacklevel)

        if log_key not in entries:
            context_list: LossyList[str] = LossyList()
            if context is not None:
                context_list.append(context)
            entries[log_key] = StructuredLogEntry(
                title=title,
                message=message,
                context=context_list,
            )
        else:
            if context is not None:
                entries[log_key].context.append(context)

    def _get_of_type(self, level: StructuredLogLevel) -> LossyList[StructuredLogEntry]:
        entries = self._entries[level]
        result: LossyList[StructuredLogEntry] = LossyList()
        for log in entries.values():
            result.append(log)
        result.set_total(entries.total_key_count())
        return result

    @property
    def warnings(self) -> LossyList[StructuredLogEntry]:
        return self._get_of_type(StructuredLogLevel.WARN)

    @property
    def failures(self) -> LossyList[StructuredLogEntry]:
        return self._get_of_type(StructuredLogLevel.ERROR)

    @property
    def infos(self) -> LossyList[StructuredLogEntry]:
        return self._get_of_type(StructuredLogLevel.INFO)


@dataclass
class SourceReport(Report):
    events_produced: int = 0
    events_produced_per_sec: int = 0

    _urns_seen: Set[str] = field(default_factory=set)
    entities: Dict[str, list] = field(default_factory=lambda: defaultdict(LossyList))
    aspects: Dict[str, Dict[str, int]] = field(
        default_factory=lambda: defaultdict(lambda: defaultdict(int))
    )
    aspect_urn_samples: Dict[str, Dict[str, LossyList[str]]] = field(
        default_factory=lambda: defaultdict(lambda: defaultdict(LossyList))
    )

    _structured_logs: StructuredLogs = field(default_factory=StructuredLogs)

    @property
    def warnings(self) -> LossyList[StructuredLogEntry]:
        return self._structured_logs.warnings

    @property
    def failures(self) -> LossyList[StructuredLogEntry]:
        return self._structured_logs.failures

    @property
    def infos(self) -> LossyList[StructuredLogEntry]:
        return self._structured_logs.infos

    def report_workunit(self, wu: WorkUnit) -> None:
        self.events_produced += 1

        if isinstance(wu, MetadataWorkUnit):
            urn = wu.get_urn()

            # Specialized entity reporting.
            if not isinstance(wu.metadata, MetadataChangeEvent):
                mcps = [wu.metadata]
            else:
                mcps = list(mcps_from_mce(wu.metadata))

            for mcp in mcps:
                entityType = mcp.entityType
                aspectName = mcp.aspectName

                if urn not in self._urns_seen:
                    self._urns_seen.add(urn)
                    self.entities[entityType].append(urn)

                if aspectName is not None:  # usually true
                    self.aspects[entityType][aspectName] += 1
                    self.aspect_urn_samples[entityType][aspectName].append(urn)
                    if isinstance(mcp.aspect, UpstreamLineageClass):
                        upstream_lineage = cast(UpstreamLineageClass, mcp.aspect)
                        if upstream_lineage.fineGrainedLineages:
                            self.aspect_urn_samples[entityType][
                                "fineGrainedLineages"
                            ].append(urn)

    def report_warning(
        self,
        message: LiteralString,
        context: Optional[str] = None,
        title: Optional[LiteralString] = None,
        exc: Optional[BaseException] = None,
    ) -> None:
        self._structured_logs.report_log(
            StructuredLogLevel.WARN, message, title, context, exc, log=False
        )

    def warning(
        self,
        message: LiteralString,
        context: Optional[str] = None,
        title: Optional[LiteralString] = None,
        exc: Optional[BaseException] = None,
    ) -> None:
        self._structured_logs.report_log(
            StructuredLogLevel.WARN, message, title, context, exc, log=True
        )

    def report_failure(
        self,
        message: LiteralString,
        context: Optional[str] = None,
        title: Optional[LiteralString] = None,
        exc: Optional[BaseException] = None,
        log: bool = True,
    ) -> None:
        self._structured_logs.report_log(
            StructuredLogLevel.ERROR, message, title, context, exc, log=log
        )

    def failure(
        self,
        message: LiteralString,
        context: Optional[str] = None,
        title: Optional[LiteralString] = None,
        exc: Optional[BaseException] = None,
        log: bool = True,
    ) -> None:
        self._structured_logs.report_log(
            StructuredLogLevel.ERROR, message, title, context, exc, log=log
        )

    def info(
        self,
        message: LiteralString,
        context: Optional[str] = None,
        title: Optional[LiteralString] = None,
        exc: Optional[BaseException] = None,
        log: bool = True,
    ) -> None:
        self._structured_logs.report_log(
            StructuredLogLevel.INFO, message, title, context, exc, log=log
        )

    @contextlib.contextmanager
    def report_exc(
        self,
        message: LiteralString,
        title: Optional[LiteralString] = None,
        context: Optional[str] = None,
        level: StructuredLogLevel = StructuredLogLevel.ERROR,
    ) -> Iterator[None]:
        # Convenience method that helps avoid boilerplate try/except blocks.
        # TODO: I'm not super happy with the naming here - it's not obvious that this
        # suppresses the exception in addition to reporting it.
        try:
            yield
        except Exception as exc:
            self._structured_logs.report_log(
                level, message=message, title=title, context=context, exc=exc
            )

    def __post_init__(self) -> None:
        self.start_time = datetime.datetime.now()
        self.running_time: datetime.timedelta = datetime.timedelta(seconds=0)

    def as_obj(self) -> dict:
        return {
            **super().as_obj(),
            # To reduce the amount of nesting, we pull these fields out of the structured log.
            "failures": Report.to_pure_python_obj(self.failures),
            "warnings": Report.to_pure_python_obj(self.warnings),
            "infos": Report.to_pure_python_obj(self.infos),
        }

    def compute_stats(self) -> None:
        duration = datetime.datetime.now() - self.start_time
        workunits_produced = self.events_produced
        if duration.total_seconds() > 0:
            self.events_produced_per_sec: int = int(
                workunits_produced / duration.total_seconds()
            )
            self.running_time = duration
        else:
            self.read_rate = 0


class CapabilityReport(BaseModel):
    """A report capturing the result of any capability evaluation"""

    capable: bool
    failure_reason: Optional[str] = None
    mitigation_message: Optional[str] = None


@dataclass
class TestConnectionReport(Report):
    internal_failure: Optional[bool] = None
    internal_failure_reason: Optional[str] = None
    basic_connectivity: Optional[CapabilityReport] = None
    capability_report: Optional[
        Dict[Union[SourceCapability, str], CapabilityReport]
    ] = None


WorkUnitType = TypeVar("WorkUnitType", bound=WorkUnit)
ExtractorConfig = TypeVar("ExtractorConfig", bound=ConfigModel)

WorkUnitProcessor = Callable[[Iterable[WorkUnitType]], Iterable[WorkUnitType]]
MetadataWorkUnitProcessor = WorkUnitProcessor[MetadataWorkUnit]


class Extractor(Generic[WorkUnitType, ExtractorConfig], Closeable, metaclass=ABCMeta):
    ctx: PipelineContext
    config: ExtractorConfig

    @classmethod
    def get_config_class(cls) -> Type[ExtractorConfig]:
        config_class = get_class_from_annotation(cls, Extractor, ConfigModel)
        assert config_class, "Extractor subclasses must define a config class"
        return cast(Type[ExtractorConfig], config_class)

    def __init__(self, config_dict: dict, ctx: PipelineContext) -> None:
        super().__init__()

        config_class = self.get_config_class()

        self.ctx = ctx
        self.config = config_class.parse_obj(config_dict)

    @abstractmethod
    def get_records(self, workunit: WorkUnitType) -> Iterable[RecordEnvelope]:
        pass


@dataclass
class Source(Closeable, metaclass=ABCMeta):
    ctx: PipelineContext

    @classmethod
    def create(cls, config_dict: dict, ctx: PipelineContext) -> "Source":
        # Technically, this method should be abstract. However, the @config_class
        # decorator automatically generates a create method at runtime if one is
        # not defined. Python still treats the class as abstract because it thinks
        # the create method is missing. To avoid the class becoming abstract, we
        # can't make this method abstract.
        raise NotImplementedError('sources must implement "create"')

    def get_workunit_processors(self) -> List[Optional[MetadataWorkUnitProcessor]]:
        """A list of functions that transforms the workunits produced by this source.
        Run in order, first in list is applied first. Be careful with order when overriding.
        """
        browse_path_processor: Optional[MetadataWorkUnitProcessor] = None
        if (
            self.ctx.pipeline_config
            and self.ctx.pipeline_config.flags.generate_browse_path_v2
        ):
            browse_path_processor = self._get_browse_path_processor(
                self.ctx.pipeline_config.flags.generate_browse_path_v2_dry_run
            )

        auto_lowercase_dataset_urns: Optional[MetadataWorkUnitProcessor] = None
        if (
            self.ctx.pipeline_config
            and self.ctx.pipeline_config.source
            and self.ctx.pipeline_config.source.config
            and (
                (
                    hasattr(
                        self.ctx.pipeline_config.source.config,
                        "convert_urns_to_lowercase",
                    )
                    and self.ctx.pipeline_config.source.config.convert_urns_to_lowercase
                )
                or (
                    hasattr(self.ctx.pipeline_config.source.config, "get")
                    and self.ctx.pipeline_config.source.config.get(
                        "convert_urns_to_lowercase"
                    )
                )
            )
        ):
            auto_lowercase_dataset_urns = auto_lowercase_urns

        return [
            auto_lowercase_dataset_urns,
            auto_status_aspect,
            auto_materialize_referenced_tags_terms,
            partial(
                auto_fix_duplicate_schema_field_paths, platform=self._infer_platform()
            ),
            browse_path_processor,
            partial(auto_workunit_reporter, self.get_report()),
        ]

    @staticmethod
    def _apply_workunit_processors(
        workunit_processors: Sequence[Optional[MetadataWorkUnitProcessor]],
        stream: Iterable[MetadataWorkUnit],
    ) -> Iterable[MetadataWorkUnit]:
        for processor in workunit_processors:
            if processor is not None:
                stream = processor(stream)
        return stream

    def get_workunits(self) -> Iterable[MetadataWorkUnit]:
        return self._apply_workunit_processors(
            self.get_workunit_processors(), self.get_workunits_internal()
        )

    def get_workunits_internal(self) -> Iterable[MetadataWorkUnit]:
        raise NotImplementedError(
            "get_workunits_internal must be implemented if get_workunits is not overriden."
        )

    def get_config(self) -> Optional[ConfigModel]:
        """Overridable method to return the config object for this source.

        Enables defining workunit processors in this class, rather than per source.
        More broadly, this method contributes to the standardization of sources,
        to promote more source-generic functionality.

        Eventually, would like to replace this call with a Protocol that requires
        a config object to be defined on each source.
        """
        return getattr(self, "config", None) or getattr(self, "source_config", None)

    @abstractmethod
    def get_report(self) -> SourceReport:
        pass

    def close(self) -> None:
        pass

    def _infer_platform(self) -> Optional[str]:
        config = self.get_config()
        return (
            getattr(config, "platform_name", None)
            or getattr(self, "platform", None)
            or getattr(config, "platform", None)
        )

    def _get_browse_path_processor(self, dry_run: bool) -> MetadataWorkUnitProcessor:
        config = self.get_config()

        platform = self._infer_platform()
        env = getattr(config, "env", None)
        browse_path_drop_dirs = [
            platform,
            platform and platform.lower(),
            env,
            env and env.lower(),
        ]

        platform_instance: Optional[str] = None
        if isinstance(config, PlatformInstanceConfigMixin) and config.platform_instance:
            platform_instance = config.platform_instance

        browse_path_processor = partial(
            auto_browse_path_v2,
            platform=platform,
            platform_instance=platform_instance,
            drop_dirs=[s for s in browse_path_drop_dirs if s is not None],
            dry_run=dry_run,
        )
        return lambda stream: browse_path_processor(stream)


class TestableSource(Source):
    @staticmethod
    @abstractmethod
    def test_connection(config_dict: dict) -> TestConnectionReport:
        raise NotImplementedError("This class does not implement this method")



import enum
from typing import Optional

from typing_extensions import TypedDict

from datahub.metadata.schema_classes import OperationTypeClass

PLATFORMS_WITH_CASE_SENSITIVE_TABLES = {
    "bigquery",
}

DIALECTS_WITH_CASE_INSENSITIVE_COLS = {
    # Column identifiers are case-insensitive in BigQuery, so we need to
    # do a normalization step beforehand to make sure it's resolved correctly.
    "bigquery",
    # Our snowflake source lowercases column identifiers, so we are forced
    # to do fuzzy (case-insensitive) resolution instead of exact resolution.
    "snowflake",
    # Teradata column names are case-insensitive.
    # A name, even when enclosed in double quotation marks, is not case sensitive. For example, CUSTOMER and Customer are the same.
    # See more below:
    # https://documentation.sas.com/doc/en/pgmsascdc/9.4_3.5/acreldb/n0ejgx4895bofnn14rlguktfx5r3.htm
    "teradata",
}
DIALECTS_WITH_DEFAULT_UPPERCASE_COLS = {
    # In some dialects, column identifiers are effectively case insensitive
    # because they are automatically converted to uppercase. Most other systems
    # automatically lowercase unquoted identifiers.
    "snowflake",
}


class QueryType(enum.Enum):
    UNKNOWN = "UNKNOWN"

    CREATE_DDL = "CREATE_DDL"
    CREATE_VIEW = "CREATE_VIEW"
    CREATE_TABLE_AS_SELECT = "CREATE_TABLE_AS_SELECT"
    CREATE_OTHER = "CREATE_OTHER"

    SELECT = "SELECT"
    INSERT = "INSERT"
    UPDATE = "UPDATE"
    DELETE = "DELETE"
    MERGE = "MERGE"

    def is_create(self) -> bool:
        return self in {
            QueryType.CREATE_DDL,
            QueryType.CREATE_VIEW,
            QueryType.CREATE_TABLE_AS_SELECT,
            QueryType.CREATE_OTHER,
        }

    def to_operation_type(self) -> Optional[str]:
        if self.is_create():
            return OperationTypeClass.CREATE

        query_to_operation_mapping = {
            QueryType.SELECT: None,
            QueryType.INSERT: OperationTypeClass.INSERT,
            QueryType.UPDATE: OperationTypeClass.UPDATE,
            QueryType.DELETE: OperationTypeClass.DELETE,
            QueryType.MERGE: OperationTypeClass.UPDATE,
        }
        return query_to_operation_mapping.get(self, OperationTypeClass.UNKNOWN)


class QueryTypeProps(TypedDict, total=False):
    kind: str  # used for create statements
    temporary: bool



from typing_extensions import Literal

from datahub.configuration.pydantic_migration_helpers import v1_ConfigModel


class SqlFilter(v1_ConfigModel):
    type: Literal["sql"]
    sql: str


DatasetFilter = SqlFilter
# class DatasetFilter(v1_ConfigModel):
#    __root__: Union[SqlFilter] = v1_Field(discriminator="type")



import json
import logging
import pathlib
import time
from typing import Any, Dict, Iterable, List, Optional, Type, Union

import duckdb

from datahub.emitter.aspect import ASPECT_MAP
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.emitter.mcp_builder import mcps_from_mce
from datahub.emitter.serialization_helper import post_json_transform
from datahub.lite.duckdb_lite_config import DuckDBLiteConfig
from datahub.lite.lite_local import (
    AutoComplete,
    Browseable,
    DataHubLiteLocal,
    PathNotFoundException,
    Searchable,
    SearchFlavor,
)
from datahub.metadata.schema_classes import (
    ChartInfoClass,
    ContainerClass,
    ContainerPropertiesClass,
    DashboardInfoClass,
    DataPlatformInstanceClass,
    DatasetPropertiesClass,
    MetadataChangeEventClass,
    SubTypesClass,
    SystemMetadataClass,
    TagPropertiesClass,
    _Aspect,
)
from datahub.utilities.urns.data_platform_urn import DataPlatformUrn
from datahub.utilities.urns.dataset_urn import DatasetUrn
from datahub.utilities.urns.urn import Urn

logger = logging.getLogger(__name__)


class DuckDBLite(DataHubLiteLocal[DuckDBLiteConfig]):
    @classmethod
    def create(cls, config_dict: dict) -> "DuckDBLite":
        config: DuckDBLiteConfig = DuckDBLiteConfig.parse_obj(config_dict)
        return DuckDBLite(config)

    def __init__(self, config: DuckDBLiteConfig) -> None:
        self.config = config
        fpath = pathlib.Path(self.config.file)
        fpath.parent.mkdir(exist_ok=True)
        self.duckdb_client = duckdb.connect(
            str(fpath), read_only=config.read_only, config=config.options
        )
        if not config.read_only:
            self._init_db()

    def _create_unique_index(
        self, index_name: str, table_name: str, columns: list
    ) -> None:
        try:
            self.duckdb_client.execute(
                f"CREATE UNIQUE INDEX {index_name} ON {table_name} ({', '.join(columns)})"
            )
        except duckdb.CatalogException as e:
            if "already exists" not in str(e).lower():
                raise

    def _init_db(self) -> None:
        self.duckdb_client.execute(
            "CREATE TABLE IF NOT EXISTS metadata_aspect_v2 "
            "(urn VARCHAR, aspect_name VARCHAR, version BIGINT, metadata JSON, system_metadata JSON, createdon BIGINT)"
        )

        self._create_unique_index(
            "aspect_idx", "metadata_aspect_v2", ["urn", "aspect_name", "version"]
        )

        self.duckdb_client.execute(
            "CREATE TABLE IF NOT EXISTS metadata_edge_v2 "
            "(src_id VARCHAR, relnship VARCHAR, dst_id VARCHAR, dst_label VARCHAR)"
        )

        self._create_unique_index(
            "edge_idx", "metadata_edge_v2", ["src_id", "relnship", "dst_id"]
        )

    def location(self) -> str:
        return self.config.file

    def destroy(self) -> None:
        fpath = pathlib.Path(self.config.file)
        fpath.unlink()

    def write(
        self,
        record: Union[
            MetadataChangeEventClass,
            MetadataChangeProposalWrapper,
        ],
    ) -> None:
        writeables: Iterable[MetadataChangeProposalWrapper]
        if isinstance(record, MetadataChangeProposalWrapper):
            writeables = [record]
        elif isinstance(record, MetadataChangeEventClass):
            writeables = mcps_from_mce(record)
        else:
            raise ValueError(
                f"DuckDBCatalog only supports MCEs and MCPs, not {type(record)}"
            )

        if not writeables:
            return

        # TODO use `with` for transaction
        self.duckdb_client.begin()
        for writeable in writeables:
            needs_write = False
            try:
                writeable_dict = writeable.to_obj(simplified_structure=True)
                exists = self.duckdb_client.execute(
                    "SELECT metadata, system_metadata FROM metadata_aspect_v2 WHERE urn = ? AND aspect_name = ? AND version = 0",
                    [writeable.entityUrn, writeable.aspectName],
                )
                max_row = exists.fetchone()
                if max_row is None:
                    new_version = 1
                    needs_write = True
                else:
                    metadata_dict = json.loads(max_row[0])  # type: ignore
                    system_metadata = json.loads(max_row[1])  # type: ignore
                    real_version = system_metadata.get("properties", {}).get(
                        "sysVersion"
                    )
                    if real_version is None:
                        max_version_row = self.duckdb_client.execute(
                            "SELECT max(version) FROM metadata_aspect_v2 WHERE urn = ? AND aspect_name = ?",
                            [writeable.entityUrn, writeable.aspectName],
                        ).fetchone()
                        real_version = max_version_row[0]  # type: ignore

                    if writeable_dict["aspect"]["json"] == metadata_dict:
                        needs_write = False
                        new_version = real_version
                    else:
                        needs_write = True
                        new_version = real_version + 1

                current_time = int(time.time() * 1000.0)
                created_on = current_time
                if (
                    writeable.systemMetadata is not None
                    and writeable.systemMetadata.lastObserved
                ):
                    created_on = writeable.systemMetadata.lastObserved

                if writeable.systemMetadata is None:
                    writeable.systemMetadata = SystemMetadataClass(
                        lastObserved=created_on, properties={}
                    )
                elif writeable.systemMetadata.lastObserved is None:
                    writeable.systemMetadata.lastObserved = created_on

                if "properties" not in writeable_dict["systemMetadata"]:
                    writeable_dict["systemMetadata"]["properties"] = {}
                writeable_dict["systemMetadata"]["properties"][
                    "sysVersion"
                ] = new_version
                if needs_write:
                    self.duckdb_client.execute(
                        query="INSERT INTO metadata_aspect_v2 VALUES (?, ?, ?, ?, ?, ?)",
                        parameters=[
                            writeable.entityUrn,
                            writeable.aspectName,
                            new_version,
                            json.dumps(writeable_dict["aspect"]["json"]),
                            json.dumps(writeable_dict["systemMetadata"]),
                            created_on,
                        ],
                    )
                    if not max_row:
                        self.duckdb_client.execute(
                            query="INSERT INTO metadata_aspect_v2 VALUES (?, ?, ?, ?, ?, ?)",
                            parameters=[
                                writeable.entityUrn,
                                writeable.aspectName,
                                0,
                                json.dumps(writeable_dict["aspect"]["json"]),
                                json.dumps(writeable_dict["systemMetadata"]),
                                created_on,
                            ],
                        )
                    else:
                        # we update the existing v0 row
                        self.duckdb_client.execute(
                            query="UPDATE metadata_aspect_v2 SET metadata = ?, system_metadata = ? WHERE urn = ? AND aspect_name = ? AND version = 0",
                            parameters=[
                                json.dumps(writeable_dict["aspect"]["json"]),
                                json.dumps(writeable_dict["systemMetadata"]),
                                writeable.entityUrn,
                                writeable.aspectName,
                            ],
                        )
                else:
                    # this is a dup, we still want to update the lastObserved timestamp
                    if not system_metadata:
                        system_metadata = {
                            "lastObserved": writeable.systemMetadata.lastObserved
                        }
                    else:
                        system_metadata[
                            "lastObserved"
                        ] = writeable.systemMetadata.lastObserved
                    self.duckdb_client.execute(
                        query="UPDATE metadata_aspect_v2 SET system_metadata = ? WHERE urn = ? AND aspect_name = ? AND version = 0",
                        parameters=[
                            json.dumps(system_metadata),
                            writeable.entityUrn,
                            writeable.aspectName,
                        ],
                    )
            except Exception as e:
                logger.error(f"Failed to write {writeable}", e)
            else:
                if needs_write:
                    assert (
                        writeable.entityUrn
                        and writeable.aspectName
                        and writeable.aspect
                    )
                    self.post_update_hook(
                        writeable.entityUrn, writeable.aspectName, writeable.aspect
                    )

        self.duckdb_client.commit()

    def list_ids(self) -> Iterable[str]:
        self.duckdb_client.execute("SELECT distinct(urn) from metadata_aspect_v2")
        for row in self.duckdb_client.fetchall():
            yield row[0]

    def get(
        self,
        id: str,
        aspects: Optional[List[str]],
        typed: bool = False,
        as_of: Optional[int] = None,
        details: Optional[bool] = False,
    ) -> Optional[Dict[str, Union[str, dict, _Aspect]]]:
        base_query = "SELECT urn, aspect_name, metadata, system_metadata from metadata_aspect_v2 WHERE urn = ?"
        if aspects:
            base_query += (
                " AND aspect_name IN (" + ",".join([f"'{x}'" for x in aspects]) + ")"
            )
        if as_of:
            base_query += f" AND createdon < '{as_of}' ORDER BY version DESC LIMIT 1"
        else:
            base_query += " AND version = 0"

        self.duckdb_client.execute(base_query, [id])
        results = self.duckdb_client.fetchall()
        result_map: Dict[str, Union[str, dict, _Aspect]] = {}
        for r in results:
            aspect_name: str = r[1]
            aspect: Union[dict, _Aspect] = json.loads(r[2])
            if typed:
                assert isinstance(aspect, dict)
                aspect = ASPECT_MAP[aspect_name].from_obj(post_json_transform(aspect))

            result_map[aspect_name] = aspect
            if details:
                system_metadata: Union[dict, SystemMetadataClass] = json.loads(r[3])
                if typed:
                    assert isinstance(system_metadata, dict)
                    system_metadata = SystemMetadataClass.from_obj(system_metadata)
                result_map[aspect_name].update({"__systemMetadata": system_metadata})  # type: ignore
        if result_map:
            result_map = {**{"urn": id}, **result_map}
            return result_map
        else:
            return None

    def search(
        self,
        query: str,
        flavor: SearchFlavor,
        aspects: List[str] = [],
        snippet: bool = True,
    ) -> Iterable[Searchable]:
        if flavor == SearchFlavor.FREE_TEXT:
            base_query = f"SELECT distinct(urn), 'urn', NULL from metadata_aspect_v2 where urn ILIKE '%{query}%' UNION SELECT urn, aspect_name, metadata from metadata_aspect_v2 where metadata->>'$.name' ILIKE '%{query}%'"
            for r in self.duckdb_client.execute(base_query).fetchall():
                yield Searchable(
                    id=r[0], aspect=r[1], snippet=r[2] if snippet else None
                )
        elif flavor == SearchFlavor.EXACT:
            base_query = f"SELECT urn, aspect_name, metadata from metadata_aspect_v2 where version = 0 AND ({query})"
            for r in self.duckdb_client.execute(base_query).fetchall():
                yield Searchable(
                    id=r[0], aspect=r[1], snippet=r[2] if snippet else None
                )
        else:
            raise Exception(f"Unhandled search flavor {flavor}")

    def remove_edge(self, src: str, relnship: str) -> None:
        try:
            self.duckdb_client.execute(
                "DELETE FROM metadata_edge_v2 WHERE src_id = ? AND relnship = ?",
                [src, relnship],
            )
        except Exception as e:
            logger.error("Failed to remove any edge", e)
        else:
            self.duckdb_client.commit()

    def add_edge(
        self,
        src: Union[Urn, str],
        relnship: str,
        dst: Union[Urn, str],
        dst_label: Optional[str] = None,
        remove_existing: bool = False,
    ) -> None:
        src_id = str(src)
        dst_id = str(dst)
        logger.debug(f"Add edge {src_id},{dst_id},{relnship},{dst_label}")
        try:
            query = "SELECT * FROM metadata_edge_v2 WHERE src_id = ? AND relnship = ?"
            params = [src_id, relnship]
            if not remove_existing:
                query += " AND dst_id = ?"
                params.append(dst_id)
            exists = self.duckdb_client.execute(query, params)
            maybe_row = exists.fetchone()
            if not maybe_row:
                self.duckdb_client.execute(
                    query="INSERT INTO metadata_edge_v2 VALUES (?, ?, ?, ?)",
                    parameters=[src_id, relnship, dst_id, dst_label],
                )
            else:
                read_dst_id = maybe_row[2]  # type: ignore
                read_dst_label = maybe_row[3]  # type: ignore
                update_fragment = ""
                update_params: List[Any] = []
                if read_dst_id != dst_id:
                    update_fragment = "dst_id = ?"
                    update_params = [dst_id]
                    where_clause = "src_id = ? AND relnship = ?"
                    where_params = [src_id, relnship]

                if read_dst_label != dst_label:
                    update_fragment = (
                        ",dst_label = ?" if update_fragment else "dst_label = ?"
                    )
                    update_params += [dst_label]
                    where_clause = "src_id = ? AND relnship = ? AND dst_id = ?"
                    where_params = [src_id, relnship, dst_id]
                    if read_dst_id != dst_id:
                        where_clause = "src_id = ? AND relnship = ?"
                        where_params = [src_id, relnship]

                if update_fragment:
                    query = f"UPDATE metadata_edge_v2 SET {update_fragment} WHERE {where_clause}"
                    params = update_params + where_params
                    self.duckdb_client.execute(
                        query=query,
                        parameters=params,
                    )
        except Exception as e:
            logger.error(
                f"Failed to write {src_id}, {relnship}, {dst_id}, {dst_label}", e
            )
            raise

        self.duckdb_client.commit()

    def ls(self, path: str) -> List[Browseable]:
        def get_id_for_name(
            name: str,
            allowed_src_ids: Optional[List[str]] = None,
            expand_search: bool = False,
        ) -> List[str]:
            if not expand_search:
                query = f"SELECT src_id from metadata_edge_v2 where dst_id ='{name}' and relnship = 'name'"
            else:
                query = f"SELECT src_id from metadata_edge_v2 where dst_id ILIKE '{name}%' and relnship = 'name'"
            results = self.duckdb_client.execute(query).fetchall()
            if not results:
                return []
            else:
                ids_returned = [r[0] for r in results]
                if allowed_src_ids:
                    allowed_src_ids_quoted = [f"'{r}'" for r in allowed_src_ids]
                    query = f"SELECT dst_id from metadata_edge_v2 where src_id IN ({','.join(allowed_src_ids_quoted)})"
                    allowed_ids = [
                        r[0] for r in self.duckdb_client.execute(query).fetchall()
                    ]
                    ids_returned = [id for id in ids_returned if id in allowed_ids]
                return ids_returned

        def resolve_name_from_id(maybe_urn: str) -> str:
            query = f"SELECT dst_id from metadata_edge_v2 where src_id = '{maybe_urn}' and relnship = 'name'"
            results = self.duckdb_client.execute(query).fetchall()
            if not results:
                return maybe_urn
            else:
                return results[0][0]

        pieces = [p for p in path.split("/") if p]

        pieces = ["__root__"] + pieces

        # very lazy walk
        in_list = [pieces[0]]
        in_list_quoted = [f"'{r}'" for r in in_list]

        for i, p in enumerate(pieces[1:]):
            ids = get_id_for_name(p, allowed_src_ids=in_list)
            if ids:
                ids_quoted = [f"'{r}'" for r in ids]
                query = f"SELECT dst_id from metadata_edge_v2 where relnship = 'child' AND src_id IN ({','.join(in_list_quoted)}) and dst_id IN ({','.join(ids_quoted)})"
            else:
                query = f"SELECT dst_id from metadata_edge_v2 where relnship = 'child' AND src_id IN ({','.join(in_list_quoted)}) and (dst_id = '{p}' OR dst_label = '{p}')"
            results = self.duckdb_client.execute(query).fetchall()
            if not results:
                ids = get_id_for_name(p, allowed_src_ids=in_list, expand_search=True)
                alternatives = None
                if ids:
                    ids_quoted = [f"'{r}'" for r in ids]
                    query = f"SELECT dst_id from metadata_edge_v2 where relnship = 'child' AND src_id IN ({','.join(in_list_quoted)}) and dst_id IN ({','.join(ids_quoted)})"
                    results = self.duckdb_client.execute(query).fetchall()
                    success_path = "/" + "/".join(pieces[1 : i + 1])
                    results_list = []
                    for r in results:
                        r_name = resolve_name_from_id(r[0])
                        results_list.append(
                            Browseable(
                                id=r[0],
                                name=r_name,
                                leaf=False,
                                parents=in_list,
                                auto_complete=AutoComplete(
                                    success_path=success_path,
                                    failed_token=p,
                                    suggested_path=f"{success_path}/{r_name}".replace(
                                        "//", "/"
                                    ),
                                ),
                            )
                        )
                    return results_list
                raise PathNotFoundException(
                    f"Path {path} not found at {p} for query: {query}, did you mean {alternatives}"
                )
            in_list = [r[0] for r in results]
            in_list_quoted = [f"'{r}'" for r in in_list]

        query = f"SELECT dst_id, dst_label from metadata_edge_v2 where src_id IN ({','.join(in_list_quoted)}) and relnship = 'child'"
        results = self.duckdb_client.execute(query).fetchall()
        if results:
            results_list = [
                Browseable(
                    parents=in_list,
                    id=r[0],
                    name=r[1] if r[1] else resolve_name_from_id(r[0]),
                )
                for r in results
            ]
            return results_list
        else:
            # this is a leaf, return the urn for the entity
            return [
                Browseable(id=r, name=pieces[-1], leaf=True)
                for r in in_list
                if r != "__root__"
            ]

    def reindex(self) -> None:
        self.duckdb_client.execute("DELETE FROM metadata_edge_v2")
        self.duckdb_client.commit()
        for urn_aspect_dict in self.get_all_entities(typed=True):
            for urn, aspect_map in urn_aspect_dict.items():
                for aspect_name, aspect_value in aspect_map.items():
                    assert isinstance(aspect_value, _Aspect)
                    self.post_update_hook(urn, aspect_name, aspect_value)
                self.global_post_update_hook(urn, aspect_map)  # type: ignore

    def get_all_entities(
        self, typed: bool = False
    ) -> Iterable[Dict[str, Union[dict, _Aspect]]]:
        query = "SELECT urn, aspect_name, metadata, system_metadata from metadata_aspect_v2 where version = 0 order by (urn, aspect_name)"
        results = self.duckdb_client.execute(query)
        aspect_map: Dict[str, _Aspect] = {}
        current_urn = None
        for r in results.fetchall():
            urn = r[0]
            aspect_name = r[1]
            aspect_payload = json.loads(r[2])
            if typed:
                assert (
                    aspect_name in ASPECT_MAP
                ), f"Missing aspect name {aspect_name} in the registry"
                try:
                    aspect_payload = ASPECT_MAP[aspect_name].from_obj(
                        post_json_transform(aspect_payload)
                    )
                except Exception as e:
                    logger.exception(
                        f"Failed to process urn: {urn}, aspect_name: {aspect_name}, metadata: {aspect_payload}",
                        exc_info=e,
                    )
                    raise

            if current_urn is None:
                current_urn = urn
            if urn != current_urn:
                if aspect_map:
                    yield {current_urn: aspect_map}
                    aspect_map = {}
                    current_urn = urn

            aspect_map[aspect_name] = aspect_payload

        if aspect_map:
            assert current_urn
            yield {current_urn: aspect_map}

    def get_all_aspects(self) -> Iterable[MetadataChangeProposalWrapper]:
        query = "SELECT urn, aspect_name, metadata, system_metadata from metadata_aspect_v2 where version = 0"
        results = self.duckdb_client.execute(query)
        for r in results.fetchall():
            urn = r[0]
            aspect_name = r[1]
            aspect_metadata = ASPECT_MAP[aspect_name].from_obj(post_json_transform(json.loads(r[2])))  # type: ignore
            system_metadata = SystemMetadataClass.from_obj(json.loads(r[3]))
            mcp = MetadataChangeProposalWrapper(
                entityUrn=urn,
                aspectName=aspect_name,
                aspect=aspect_metadata,
                systemMetadata=system_metadata,
            )
            yield mcp

    def close(self) -> None:
        self.reindex()
        self.duckdb_client.close()

    def get_category_from_platform(self, data_platform_urn: DataPlatformUrn) -> Urn:
        # TODO: Use the dataplatforms.json resource to auto-gen this
        category_to_platform_map = {
            "bi_tools": ["looker", "tableau", "powerbi", "superset"],
            "databases": [
                "mysql",
                "snowflake",
                "postgres",
                "bigquery",
                "redshift",
                "clickhouse",
            ],
            "data_lakes": [
                "s3",
                "hdfs",
                "delta-lake",
                "glue",
                "hive",
                "presto",
                "iceberg",
                "trino",
            ],
            "streaming": ["kafka"],
            "orchestrators": ["airflow", "spark"],
            "data_movers": ["kafka-connect", "nifi"],
            "transformation_tools": ["dbt"],
            "data_quality": ["great-expectations"],
        }
        for k, v in category_to_platform_map.items():
            if data_platform_urn.get_entity_id_as_string() in v:
                return Urn(entity_type="systemNode", entity_id=[k])

        logger.debug(
            f"Failed to find category for platform {data_platform_urn}, mapping to generic data_platform"
        )
        return Urn(entity_type="systemNode", entity_id=["data_platforms"])

    def global_post_update_hook(
        self, entity_urn: str, aspect_map: Dict[str, _Aspect]
    ) -> None:
        def pluralize(noun: str) -> str:
            return noun.lower() + "s"

        def get_typed_aspect(
            aspect_map: Dict[str, _Aspect], aspect_type: Type[_Aspect]
        ) -> Optional[_Aspect]:
            aspect_names = [k for k, v in ASPECT_MAP.items() if v == aspect_type]
            if aspect_names:
                return aspect_map.get(aspect_names[0])
            raise Exception(
                f"Unable to locate aspect type {aspect_type} in the registry"
            )

        if not entity_urn:
            logger.error(f"Bad input {entity_urn}: {aspect_map}")

        container: Optional[ContainerClass] = get_typed_aspect(  # type: ignore
            aspect_map, ContainerClass
        )  # type: ignore
        subtypes: Optional[SubTypesClass] = get_typed_aspect(aspect_map, SubTypesClass)  # type: ignore
        dpi: Optional[DataPlatformInstanceClass] = get_typed_aspect(  # type: ignore
            aspect_map, DataPlatformInstanceClass
        )  # type: ignore

        needs_platform = Urn.create_from_string(entity_urn).get_type() in [
            "dataset",
            "container",
            "chart",
            "dashboard",
            "dataFlow",
            "dataJob",
        ]
        entity_urn_parsed = Urn.create_from_string(entity_urn)
        if entity_urn_parsed.get_type() in ["dataFlow", "dataJob"]:
            self.add_edge(
                entity_urn,
                "name",
                entity_urn_parsed.get_entity_id()[1],
                remove_existing=True,
            )

        if not container and needs_platform:
            # this is a top-level entity
            if not dpi:
                logger.debug(f"No data platform instance for {entity_urn}")
                maybe_parent_urn = Urn.create_from_string(entity_urn).get_entity_id()[0]
                needs_dpi = False
                if maybe_parent_urn.startswith(Urn.URN_PREFIX):
                    parent_urn = maybe_parent_urn
                    if (
                        Urn.create_from_string(maybe_parent_urn).get_type()
                        == "dataPlatform"
                    ):
                        data_platform_urn = DataPlatformUrn.create_from_string(
                            maybe_parent_urn
                        )
                        needs_dpi = True
                else:
                    data_platform_urn = DataPlatformUrn.create_from_id(maybe_parent_urn)
                    needs_dpi = True

                if needs_dpi:
                    data_platform_instance = "default"
                    data_platform_instance_urn = Urn(
                        entity_type="dataPlatformInstance",
                        entity_id=[str(data_platform_urn), data_platform_instance],
                    )
                    try:
                        self._create_edges_from_data_platform_instance(
                            data_platform_instance_urn
                        )
                    except Exception as e:
                        logger.error(f"Failed to generate edges entity {entity_urn}", e)
                    parent_urn = str(data_platform_instance_urn)
            else:
                data_platform_urn = DataPlatformUrn.create_from_string(dpi.platform)
                data_platform_instance = dpi.instance or "default"
                data_platform_instance_urn = Urn(
                    entity_type="dataPlatformInstance",
                    entity_id=[str(data_platform_urn), data_platform_instance],
                )
                parent_urn = str(data_platform_instance_urn)
        elif container:
            parent_urn = container.container
        else:
            parent_urn = "__root__"

        types = (
            subtypes.typeNames
            if subtypes
            else [Urn.create_from_string(entity_urn).get_type()]
        )
        for t in types:
            type_urn = Urn(entity_type="systemNode", entity_id=[parent_urn, t])
            self.add_edge(parent_urn, "child", type_urn)
            self.add_edge(type_urn, "child", entity_urn)
            self.add_edge(type_urn, "name", pluralize(t), remove_existing=True)

    def _create_edges_from_data_platform_instance(
        self, data_platform_instance_urn: Urn
    ) -> None:
        data_platform_urn = DataPlatformUrn.create_from_string(
            data_platform_instance_urn.get_entity_id()[0]
        )
        data_platform_instances_urn = Urn(
            entity_type="systemNode", entity_id=[str(data_platform_urn), "instances"]
        )

        data_platform_category = self.get_category_from_platform(data_platform_urn)
        # /<data_platform_category>/

        self.add_edge("__root__", "child", str(data_platform_category))
        self.add_edge(
            str(data_platform_category),
            "name",
            data_platform_category.get_entity_id_as_string(),
            remove_existing=True,
        )
        # /<data_platform_category>/<data_platform>
        self.add_edge(
            str(data_platform_category),
            "child",
            str(data_platform_urn),
            data_platform_urn.get_entity_id_as_string(),
        )
        self.add_edge(
            data_platform_urn,
            "name",
            data_platform_urn.get_entity_id_as_string(),
            remove_existing=True,
        )
        # /<data_platform_category>/<data_platform>/instances
        self.add_edge(str(data_platform_urn), "child", str(data_platform_instances_urn))
        self.add_edge(
            str(data_platform_instances_urn), "name", "instances", remove_existing=True
        )
        # /<data_platform_category>/<data_platform>/instances/<instance_name>
        self.add_edge(
            str(data_platform_instances_urn),
            "child",
            str(data_platform_instance_urn),
            data_platform_instance_urn.get_entity_id()[-1],
        )

    def post_update_hook(
        self, entity_urn: str, aspect_name: str, aspect: _Aspect
    ) -> None:
        if isinstance(aspect, DatasetPropertiesClass):
            dp: DatasetPropertiesClass = aspect
            if dp.name:
                specific_urn = DatasetUrn.create_from_string(entity_urn)
                if (
                    specific_urn.get_data_platform_urn().get_entity_id_as_string()
                    == "looker"
                ):
                    # Looker dataset urns (views and explores) need special handling
                    dataset_id = specific_urn.get_entity_id()[1]
                    self.add_edge(
                        entity_urn,
                        "name",
                        dataset_id.replace(".explore", "").replace(".view", ""),
                        remove_existing=True,
                    )
                else:
                    self.add_edge(entity_urn, "name", dp.name, remove_existing=True)
        elif isinstance(aspect, ContainerPropertiesClass):
            cp: ContainerPropertiesClass = aspect
            self.add_edge(entity_urn, "name", cp.name, remove_existing=True)
        elif isinstance(aspect, DataPlatformInstanceClass):
            dpi: DataPlatformInstanceClass = aspect
            data_platform_urn = DataPlatformUrn.create_from_string(dpi.platform)
            data_platform_instance = dpi.instance or "default"
            data_platform_instance_urn = Urn(
                entity_type="dataPlatformInstance",
                entity_id=[str(data_platform_urn), data_platform_instance],
            )
            self._create_edges_from_data_platform_instance(data_platform_instance_urn)
        elif isinstance(aspect, ChartInfoClass):
            urn = Urn.create_from_string(entity_urn)
            self.add_edge(
                entity_urn,
                "name",
                aspect.title + f" ({urn.get_entity_id()[-1]})",
                remove_existing=True,
            )
        elif isinstance(aspect, DashboardInfoClass):
            urn = Urn.create_from_string(entity_urn)
            self.add_edge(
                entity_urn,
                "name",
                aspect.title + f" ({urn.get_entity_id()[-1]})",
                remove_existing=True,
            )
        elif isinstance(aspect, TagPropertiesClass):
            self.add_edge(entity_urn, "name", aspect.name, remove_existing=True)



from typing import List

from pydantic import Field

from datahub.configuration.common import ConfigModel
from datahub.ingestion.source.data_lake_common.path_spec import PathSpec


class PathSpecsConfigMixin(ConfigModel):

    path_specs: List[PathSpec] = Field(
        description="List of PathSpec. See [below](#path-spec) the details about PathSpec"
    )



import json
import logging
import time
from datetime import datetime
from enum import Enum
from typing import Any, Dict, Iterable, List, Optional

import requests
from pydantic import Field, validator
from simple_salesforce import Salesforce
from simple_salesforce.exceptions import SalesforceAuthenticationFailed

import datahub.emitter.mce_builder as builder
from datahub.configuration.common import (
    AllowDenyPattern,
    ConfigModel,
    ConfigurationError,
)
from datahub.configuration.source_common import DatasetSourceConfigMixin
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.emitter.mcp_builder import add_domain_to_entity_wu
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SourceCapability,
    SupportStatus,
    capability,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import Source, SourceReport
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.common.subtypes import DatasetSubTypes
from datahub.ingestion.source_config.operation_config import (
    OperationConfig,
    is_profiling_enabled,
)
from datahub.metadata.schema_classes import (
    AuditStampClass,
    BooleanTypeClass,
    BytesTypeClass,
    DataPlatformInstanceClass,
    DatasetProfileClass,
    DatasetPropertiesClass,
    DateTypeClass,
    EnumTypeClass,
    ForeignKeyConstraintClass,
    GlobalTagsClass,
    NullTypeClass,
    NumberTypeClass,
    OperationClass,
    OperationTypeClass,
    OtherSchemaClass,
    RecordTypeClass,
    SchemaFieldClass,
    SchemaFieldDataTypeClass,
    SchemaMetadataClass,
    StringTypeClass,
    SubTypesClass,
    TagAssociationClass,
)
from datahub.utilities import config_clean

logger = logging.getLogger(__name__)


class SalesforceAuthType(Enum):
    USERNAME_PASSWORD = "USERNAME_PASSWORD"
    DIRECT_ACCESS_TOKEN = "DIRECT_ACCESS_TOKEN"
    JSON_WEB_TOKEN = "JSON_WEB_TOKEN"


class SalesforceProfilingConfig(ConfigModel):
    enabled: bool = Field(
        default=False,
        description="Whether profiling should be done. Supports only table-level profiling at this stage",
    )
    operation_config: OperationConfig = Field(
        default_factory=OperationConfig,
        description="Experimental feature. To specify operation configs.",
    )

    # TODO - support field level profiling


class SalesforceConfig(DatasetSourceConfigMixin):
    platform: str = "salesforce"

    auth: SalesforceAuthType = SalesforceAuthType.USERNAME_PASSWORD

    # Username, Password Auth
    username: Optional[str] = Field(description="Salesforce username")
    password: Optional[str] = Field(description="Password for Salesforce user")
    consumer_key: Optional[str] = Field(
        description="Consumer key for Salesforce JSON web token access"
    )
    private_key: Optional[str] = Field(
        description="Private key as a string for Salesforce JSON web token access"
    )
    security_token: Optional[str] = Field(
        description="Security token for Salesforce username"
    )
    # client_id, client_secret not required

    # Direct - Instance URL, Access Token Auth
    instance_url: Optional[str] = Field(
        description="Salesforce instance url. e.g. https://MyDomainName.my.salesforce.com"
    )
    # Flag to indicate whether the instance is production or sandbox
    is_sandbox: bool = Field(
        default=False, description="Connect to Sandbox instance of your Salesforce"
    )
    access_token: Optional[str] = Field(description="Access token for instance url")

    ingest_tags: Optional[bool] = Field(
        default=False,
        description="Ingest Tags from source. This will override Tags entered from UI",
    )

    object_pattern: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description="Regex patterns for Salesforce objects to filter in ingestion.",
    )
    domain: Dict[str, AllowDenyPattern] = Field(
        default=dict(),
        description='Regex patterns for tables/schemas to describe domain_key domain key (domain_key can be any string like "sales".) There can be multiple domain keys specified.',
    )
    api_version: Optional[str] = Field(
        description="If specified, overrides default version used by the Salesforce package. Example value: '59.0'"
    )

    profiling: SalesforceProfilingConfig = SalesforceProfilingConfig()

    profile_pattern: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description="Regex patterns for profiles to filter in ingestion, allowed by the `object_pattern`.",
    )

    def is_profiling_enabled(self) -> bool:
        return self.profiling.enabled and is_profiling_enabled(
            self.profiling.operation_config
        )

    @validator("instance_url")
    def remove_trailing_slash(cls, v):
        return config_clean.remove_trailing_slashes(v)


class SalesforceSourceReport(SourceReport):
    filtered: List[str] = []

    def report_dropped(self, ent_name: str) -> None:
        self.filtered.append(ent_name)


# https://developer.salesforce.com/docs/atlas.en-us.api.meta/api/sforce_api_calls_describesObjects_describesObjectresult.htm#FieldType
FIELD_TYPE_MAPPING = {
    "string": StringTypeClass,
    "boolean": BooleanTypeClass,
    "int": NumberTypeClass,
    "integer": NumberTypeClass,
    "long": NumberTypeClass,
    "double": NumberTypeClass,
    "date": DateTypeClass,
    "datetime": DateTypeClass,
    "time": DateTypeClass,
    "id": StringTypeClass,  # Primary Key
    "picklist": EnumTypeClass,
    "address": RecordTypeClass,
    "location": RecordTypeClass,
    "reference": StringTypeClass,  # Foreign Key
    "currency": NumberTypeClass,
    "textarea": StringTypeClass,
    "percent": NumberTypeClass,
    "phone": StringTypeClass,
    "url": StringTypeClass,
    "email": StringTypeClass,
    "combobox": StringTypeClass,
    "multipicklist": StringTypeClass,
    "base64": BytesTypeClass,
    "anyType": NullTypeClass,
    "encryptedstring": StringTypeClass,
}


@platform_name("Salesforce")
@config_class(SalesforceConfig)
@support_status(SupportStatus.INCUBATING)
@capability(
    capability_name=SourceCapability.PLATFORM_INSTANCE,
    description="Can be equivalent to Salesforce organization",
)
@capability(
    capability_name=SourceCapability.DOMAINS,
    description="Supported via the `domain` config field",
)
@capability(
    capability_name=SourceCapability.DATA_PROFILING,
    description="Only table level profiling is supported via `profiling.enabled` config field",
)
@capability(
    capability_name=SourceCapability.DELETION_DETECTION,
    description="Not supported yet",
    supported=False,
)
@capability(
    capability_name=SourceCapability.SCHEMA_METADATA,
    description="Enabled by default",
)
@capability(
    capability_name=SourceCapability.TAGS,
    description="Enabled by default",
)
class SalesforceSource(Source):
    base_url: str
    config: SalesforceConfig
    report: SalesforceSourceReport
    session: requests.Session
    sf: Salesforce
    fieldCounts: Dict[str, int]

    def __init__(self, config: SalesforceConfig, ctx: PipelineContext) -> None:
        super().__init__(ctx)
        self.config = config
        self.report = SalesforceSourceReport()
        self.session = requests.Session()
        self.platform: str = "salesforce"
        self.fieldCounts = {}
        common_args: Dict[str, Any] = {
            "domain": "test" if self.config.is_sandbox else None,
            "session": self.session,
        }
        if self.config.api_version:
            common_args["version"] = self.config.api_version

        try:
            if self.config.auth is SalesforceAuthType.DIRECT_ACCESS_TOKEN:
                logger.debug("Access Token Provided in Config")
                assert (
                    self.config.access_token is not None
                ), "Config access_token is required for DIRECT_ACCESS_TOKEN auth"
                assert (
                    self.config.instance_url is not None
                ), "Config instance_url is required for DIRECT_ACCESS_TOKEN auth"

                self.sf = Salesforce(
                    instance_url=self.config.instance_url,
                    session_id=self.config.access_token,
                    **common_args,
                )
            elif self.config.auth is SalesforceAuthType.USERNAME_PASSWORD:
                logger.debug("Username/Password Provided in Config")
                assert (
                    self.config.username is not None
                ), "Config username is required for USERNAME_PASSWORD auth"
                assert (
                    self.config.password is not None
                ), "Config password is required for USERNAME_PASSWORD auth"
                assert (
                    self.config.security_token is not None
                ), "Config security_token is required for USERNAME_PASSWORD auth"

                self.sf = Salesforce(
                    username=self.config.username,
                    password=self.config.password,
                    security_token=self.config.security_token,
                    **common_args,
                )

            elif self.config.auth is SalesforceAuthType.JSON_WEB_TOKEN:
                logger.debug("Json Web Token provided in the config")
                assert (
                    self.config.username is not None
                ), "Config username is required for JSON_WEB_TOKEN auth"
                assert (
                    self.config.consumer_key is not None
                ), "Config consumer_key is required for JSON_WEB_TOKEN auth"
                assert (
                    self.config.private_key is not None
                ), "Config private_key is required for JSON_WEB_TOKEN auth"

                self.sf = Salesforce(
                    username=self.config.username,
                    consumer_key=self.config.consumer_key,
                    privatekey=self.config.private_key,
                    **common_args,
                )

        except SalesforceAuthenticationFailed as e:
            logger.error(e)
            if "API_CURRENTLY_DISABLED" in str(e):
                # https://help.salesforce.com/s/articleView?id=001473830&type=1
                error = "Salesforce login failed. Please make sure user has API Enabled Access."
            else:
                error = "Salesforce login failed. Please verify your credentials."
                if (
                    self.config.instance_url
                    and "sandbox" in self.config.instance_url.lower()
                ):
                    error += "Please set `is_sandbox: True` in recipe if this is sandbox account."
            raise ConfigurationError(error) from e

        if not self.config.api_version:
            # List all REST API versions and use latest one
            versions_url = "https://{instance}/services/data/".format(
                instance=self.sf.sf_instance,
            )
            versions_response = self.sf._call_salesforce("GET", versions_url).json()
            latest_version = versions_response[-1]
            version = latest_version["version"]
            # we could avoid setting the version like below (after the Salesforce object has been already initiated
            # above), since, according to the docs:
            # https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/dome_versions.htm
            # we don't need to be authenticated to list the versions (so we could perform this call before even
            # authenticating)
            self.sf.sf_version = version

        self.base_url = "https://{instance}/services/data/v{sf_version}/".format(
            instance=self.sf.sf_instance, sf_version=self.sf.sf_version
        )

        logger.debug(
            "Using Salesforce REST API version: {version}".format(
                version=self.sf.sf_version
            )
        )

    def get_workunits_internal(self) -> Iterable[MetadataWorkUnit]:
        try:
            sObjects = self.get_salesforce_objects()
        except Exception as e:
            if "sObject type 'EntityDefinition' is not supported." in str(e):
                # https://developer.salesforce.com/docs/atlas.en-us.api_tooling.meta/api_tooling/tooling_api_objects_entitydefinition.htm
                raise ConfigurationError(
                    "Salesforce EntityDefinition query failed. "
                    "Please verify if user has 'View Setup and Configuration' permission."
                ) from e
            raise e
        else:
            for sObject in sObjects:
                yield from self.get_salesforce_object_workunits(sObject)

    def get_salesforce_object_workunits(
        self, sObject: dict
    ) -> Iterable[MetadataWorkUnit]:
        sObjectName = sObject["QualifiedApiName"]

        if not self.config.object_pattern.allowed(sObjectName):
            self.report.report_dropped(sObjectName)
            logger.debug(
                "Skipping {sObject}, as it is not allowed by object_pattern".format(
                    sObject=sObjectName
                )
            )
            return

        datasetUrn = builder.make_dataset_urn_with_platform_instance(
            self.platform,
            sObjectName,
            self.config.platform_instance,
            self.config.env,
        )

        customObject = {}
        if sObjectName.endswith("__c"):  # Is Custom Object
            customObject = self.get_custom_object_details(sObject["DeveloperName"])

            # Table Created, LastModified is available for Custom Object
            yield from self.get_operation_workunit(customObject, datasetUrn)

        yield self.get_properties_workunit(sObject, customObject, datasetUrn)

        yield from self.get_schema_metadata_workunit(
            sObjectName, sObject, customObject, datasetUrn
        )

        yield self.get_subtypes_workunit(sObjectName, datasetUrn)

        if self.config.platform_instance is not None:
            yield self.get_platform_instance_workunit(datasetUrn)

        if self.config.domain is not None:
            yield from self.get_domain_workunit(sObjectName, datasetUrn)

        if self.config.is_profiling_enabled() and self.config.profile_pattern.allowed(
            sObjectName
        ):
            yield from self.get_profile_workunit(sObjectName, datasetUrn)

    def get_custom_object_details(self, sObjectDeveloperName: str) -> dict:
        customObject = {}
        query_url = (
            self.base_url
            + "tooling/query/?q=SELECT Description, Language, ManageableState, "
            + "CreatedDate, CreatedBy.Username, LastModifiedDate, LastModifiedBy.Username "
            + f"FROM CustomObject where DeveloperName='{sObjectDeveloperName}'"
        )
        custom_objects_response = self.sf._call_salesforce("GET", query_url).json()
        if len(custom_objects_response["records"]) > 0:
            logger.debug("Salesforce CustomObject query returned with details")
            customObject = custom_objects_response["records"][0]
        return customObject

    def get_salesforce_objects(self) -> List:
        # Using Describe Global REST API returns many more objects than required.
        # Response does not have the attribute ("customizable") that can be used
        # to filter out entities not on ObjectManager UI. Hence SOQL on EntityDefinition
        # object is used instead, as suggested by salesforce support.

        query_url = (
            self.base_url
            + "tooling/query/?q=SELECT DurableId,QualifiedApiName,DeveloperName,"
            + "Label,PluralLabel,InternalSharingModel,ExternalSharingModel,DeploymentStatus "
            + "FROM EntityDefinition WHERE IsCustomizable = true"
        )
        entities_response = self.sf._call_salesforce("GET", query_url).json()
        logger.debug(
            "Salesforce EntityDefinition query returned {count} sObjects".format(
                count=len(entities_response["records"])
            )
        )
        return entities_response["records"]

    def get_domain_workunit(
        self, dataset_name: str, datasetUrn: str
    ) -> Iterable[MetadataWorkUnit]:
        domain_urn: Optional[str] = None

        for domain, pattern in self.config.domain.items():
            if pattern.allowed(dataset_name):
                domain_urn = builder.make_domain_urn(domain)

        if domain_urn:
            yield from add_domain_to_entity_wu(
                domain_urn=domain_urn, entity_urn=datasetUrn
            )

    def get_platform_instance_workunit(self, datasetUrn: str) -> MetadataWorkUnit:
        dataPlatformInstance = DataPlatformInstanceClass(
            builder.make_data_platform_urn(self.platform),
            instance=builder.make_dataplatform_instance_urn(
                self.platform, self.config.platform_instance  # type:ignore
            ),
        )

        return MetadataChangeProposalWrapper(
            entityUrn=datasetUrn, aspect=dataPlatformInstance
        ).as_workunit()

    def get_operation_workunit(
        self, customObject: dict, datasetUrn: str
    ) -> Iterable[MetadataWorkUnit]:
        reported_time: int = int(time.time() * 1000)

        if customObject.get("CreatedBy") and customObject.get("CreatedDate"):
            timestamp = self.get_time_from_salesforce_timestamp(
                customObject["CreatedDate"]
            )
            operation = OperationClass(
                timestampMillis=reported_time,
                operationType=OperationTypeClass.CREATE,
                lastUpdatedTimestamp=timestamp,
                actor=builder.make_user_urn(customObject["CreatedBy"]["Username"]),
            )

            yield MetadataChangeProposalWrapper(
                entityUrn=datasetUrn, aspect=operation
            ).as_workunit()

            # Note - Object Level LastModified captures changes at table level metadata e.g. table
            # description and does NOT capture field level metadata e.g. new field added, existing
            # field updated

            if customObject.get("LastModifiedBy") and customObject.get(
                "LastModifiedDate"
            ):
                timestamp = self.get_time_from_salesforce_timestamp(
                    customObject["LastModifiedDate"]
                )
                operation = OperationClass(
                    timestampMillis=reported_time,
                    operationType=OperationTypeClass.ALTER,
                    lastUpdatedTimestamp=timestamp,
                    actor=builder.make_user_urn(
                        customObject["LastModifiedBy"]["Username"]
                    ),
                )
                yield MetadataChangeProposalWrapper(
                    entityUrn=datasetUrn, aspect=operation
                ).as_workunit()

    def get_time_from_salesforce_timestamp(self, date: str) -> int:
        return round(
            datetime.strptime(date, "%Y-%m-%dT%H:%M:%S.%f%z").timestamp() * 1000
        )

    def get_properties_workunit(
        self, sObject: dict, customObject: Dict[str, str], datasetUrn: str
    ) -> MetadataWorkUnit:
        propertyLabels = {
            # from EntityDefinition
            "DurableId": "Durable Id",
            "DeveloperName": "Developer Name",
            "QualifiedApiName": "Qualified API Name",
            "Label": "Label",
            "PluralLabel": "Plural Label",
            "InternalSharingModel": "Internal Sharing Model",
            "ExternalSharingModel": "External Sharing Model",
            # from CustomObject
            "ManageableState": "Manageable State",
            "Language": "Language",
        }

        sObjectProperties = {
            propertyLabels[k]: str(v)
            for k, v in sObject.items()
            if k in propertyLabels and v is not None
        }
        sObjectProperties.update(
            {
                propertyLabels[k]: str(v)
                for k, v in customObject.items()
                if k in propertyLabels and v is not None
            }
        )

        datasetProperties = DatasetPropertiesClass(
            name=sObject["Label"],
            description=customObject.get("Description"),
            customProperties=sObjectProperties,
        )
        return MetadataChangeProposalWrapper(
            entityUrn=datasetUrn, aspect=datasetProperties
        ).as_workunit()

    def get_subtypes_workunit(
        self, sObjectName: str, datasetUrn: str
    ) -> MetadataWorkUnit:
        subtypes: List[str] = []
        if sObjectName.endswith("__c"):
            subtypes.append(DatasetSubTypes.SALESFORCE_CUSTOM_OBJECT)
        else:
            subtypes.append(DatasetSubTypes.SALESFORCE_STANDARD_OBJECT)

        return MetadataChangeProposalWrapper(
            entityUrn=datasetUrn, aspect=SubTypesClass(typeNames=subtypes)
        ).as_workunit()

    def get_profile_workunit(
        self, sObjectName: str, datasetUrn: str
    ) -> Iterable[MetadataWorkUnit]:
        # Here approximate record counts as returned by recordCount API are used as rowCount
        # In future, count() SOQL query may be used instead, if required, might be more expensive
        sObject_records_count_url = (
            f"{self.base_url}limits/recordCount?sObjects={sObjectName}"
        )

        sObject_record_count_response = self.sf._call_salesforce(
            "GET", sObject_records_count_url
        ).json()

        logger.debug(
            "Received Salesforce {sObject} record count response".format(
                sObject=sObjectName
            )
        )

        for entry in sObject_record_count_response.get("sObjects", []):
            datasetProfile = DatasetProfileClass(
                timestampMillis=int(time.time() * 1000),
                rowCount=entry["count"],
                columnCount=self.fieldCounts[sObjectName],
            )
            yield MetadataChangeProposalWrapper(
                entityUrn=datasetUrn, aspect=datasetProfile
            ).as_workunit()

    # Here field description is created from label, description and inlineHelpText
    def _get_field_description(self, field: dict, customField: dict) -> str:
        if "Label" not in field or field["Label"] is None:
            desc = ""
        elif field["Label"].startswith("#"):
            desc = "\\" + field["Label"]
        else:
            desc = field["Label"]

        text = field.get("FieldDefinition", {}).get("Description", None)
        if text:
            prefix = "\\" if text.startswith("#") else ""
            desc += f"\n\n{prefix}{text}"

        text = field.get("InlineHelpText", None)
        if text:
            prefix = "\\" if text.startswith("#") else ""
            desc += f"\n\n{prefix}{text}"

        return desc

    # Here jsonProps is used to add additional salesforce field level properties.
    def _get_field_json_props(self, field: dict, customField: dict) -> str:
        jsonProps = {}

        if field.get("IsUnique"):
            jsonProps["IsUnique"] = True

        return json.dumps(jsonProps)

    def _get_schema_field(
        self,
        sObjectName: str,
        fieldName: str,
        fieldType: str,
        field: dict,
        customField: dict,
    ) -> SchemaFieldClass:
        fieldPath = fieldName

        TypeClass = FIELD_TYPE_MAPPING.get(fieldType)
        if TypeClass is None:
            self.report.warning(
                message="Unable to map field type to metadata schema",
                context=f"{fieldType} for {fieldName} of {sObjectName}",
            )
            TypeClass = NullTypeClass

        fieldTags: List[str] = self.get_field_tags(fieldName, field)

        description = self._get_field_description(field, customField)

        schemaField = SchemaFieldClass(
            fieldPath=fieldPath,
            type=SchemaFieldDataTypeClass(type=TypeClass()),  # type:ignore
            description=description,
            # nativeDataType is set to data type shown on salesforce user interface,
            # not the corresponding API data type names.
            nativeDataType=field["FieldDefinition"]["DataType"],
            nullable=field["IsNillable"],
            globalTags=get_tags(fieldTags) if self.config.ingest_tags else None,
            jsonProps=self._get_field_json_props(field, customField),
        )

        # Created and LastModified Date and Actor are available for Custom Fields only
        if customField.get("CreatedDate") and customField.get("CreatedBy"):
            schemaField.created = self.get_audit_stamp(
                customField["CreatedDate"], customField["CreatedBy"]["Username"]
            )
        if customField.get("LastModifiedDate") and customField.get("LastModifiedBy"):
            schemaField.lastModified = self.get_audit_stamp(
                customField["LastModifiedDate"],
                customField["LastModifiedBy"]["Username"],
            )

        return schemaField

    def get_field_tags(self, fieldName: str, field: dict) -> List[str]:
        fieldTags: List[str] = []

        if fieldName.endswith("__c"):
            fieldTags.append("Custom")

        # https://developer.salesforce.com/docs/atlas.en-us.object_reference.meta/object_reference/system_fields.htm
        sfSystemFields = [
            "Id",
            "IsDeleted",
            "CreatedById",
            "CreatedDate",
            "LastModifiedById",
            "LastModifiedDate",
            "SystemModstamp",
        ]

        if fieldName in sfSystemFields:
            fieldTags.append("SystemField")

        if field["FieldDefinition"]["ComplianceGroup"] is not None:
            # CCPA, COPPA, GDPR, HIPAA, PCI, PersonalInfo, PII
            fieldTags.extend(
                iter(field["FieldDefinition"]["ComplianceGroup"].split(";"))
            )
        return fieldTags

    def get_audit_stamp(self, date: str, username: str) -> AuditStampClass:
        return AuditStampClass(
            time=self.get_time_from_salesforce_timestamp(date),
            actor=builder.make_user_urn(username),
        )

    def get_schema_metadata_workunit(
        self, sObjectName: str, sObject: dict, customObject: dict, datasetUrn: str
    ) -> Iterable[MetadataWorkUnit]:
        sObject_fields_query_url = (
            self.base_url
            + "tooling/query?q=SELECT "
            + "QualifiedApiName,DeveloperName,Label, FieldDefinition.DataType, DataType,"
            + "FieldDefinition.LastModifiedDate, FieldDefinition.LastModifiedBy.Username,"
            + "Precision, Scale, Length, Digits ,FieldDefinition.IsIndexed, IsUnique,"
            + "IsCompound, IsComponent, ReferenceTo, FieldDefinition.ComplianceGroup,"
            + "RelationshipName, IsNillable, FieldDefinition.Description, InlineHelpText "
            + "FROM EntityParticle WHERE EntityDefinitionId='{}'".format(
                sObject["DurableId"]
            )
        )

        sObject_fields_response = self.sf._call_salesforce(
            "GET", sObject_fields_query_url
        ).json()

        logger.debug(f"Received Salesforce {sObjectName} fields response")

        sObject_custom_fields_query_url = (
            self.base_url
            + "tooling/query?q=SELECT "
            + "DeveloperName,CreatedDate,CreatedBy.Username,InlineHelpText,"
            + "LastModifiedDate,LastModifiedBy.Username "
            + "FROM CustomField WHERE EntityDefinitionId='{}'".format(
                sObject["DurableId"]
            )
        )

        customFields: Dict[str, Dict] = {}
        try:
            sObject_custom_fields_response = self.sf._call_salesforce(
                "GET", sObject_custom_fields_query_url
            ).json()

            logger.debug(
                "Received Salesforce {sObject} custom fields response".format(
                    sObject=sObjectName
                )
            )

        except Exception as e:
            error = "Salesforce CustomField query failed. "
            if "sObject type 'CustomField' is not supported." in str(e):
                # https://github.com/afawcett/apex-toolingapi/issues/19
                error += "Please verify if user has 'View All Data' permission."

            self.report.warning(message=error, exc=e)
        else:
            customFields = {
                record["DeveloperName"]: record
                for record in sObject_custom_fields_response["records"]
            }

        fields: List[SchemaFieldClass] = []
        primaryKeys: List[str] = []
        foreignKeys: List[ForeignKeyConstraintClass] = []

        for field in sObject_fields_response["records"]:
            customField = customFields.get(field["DeveloperName"], {})

            fieldName = field["QualifiedApiName"]
            fieldType = field["DataType"]

            # Skip compound fields. All Leaf fields are ingested instead.
            if fieldType in ("address", "location"):
                continue

            schemaField: SchemaFieldClass = self._get_schema_field(
                sObjectName, fieldName, fieldType, field, customField
            )
            fields.append(schemaField)

            if fieldType == "id":
                primaryKeys.append(fieldName)

            if (
                fieldType == "reference"
                and field["ReferenceTo"]["referenceTo"] is not None
            ):
                foreignKeys.extend(
                    list(self.get_foreign_keys_from_field(fieldName, field, datasetUrn))
                )

        schemaMetadata = SchemaMetadataClass(
            schemaName="",
            platform=builder.make_data_platform_urn(self.platform),
            version=0,
            hash="",
            platformSchema=OtherSchemaClass(rawSchema=""),
            fields=fields,
            primaryKeys=primaryKeys,
            foreignKeys=foreignKeys or None,
        )

        # Created Date and Actor are available for Custom Object only
        if customObject.get("CreatedDate") and customObject.get("CreatedBy"):
            schemaMetadata.created = self.get_audit_stamp(
                customObject["CreatedDate"], customObject["CreatedBy"]["Username"]
            )
        self.fieldCounts[sObjectName] = len(fields)

        yield MetadataChangeProposalWrapper(
            entityUrn=datasetUrn, aspect=schemaMetadata
        ).as_workunit()

    def get_foreign_keys_from_field(
        self, fieldName: str, field: dict, datasetUrn: str
    ) -> Iterable[ForeignKeyConstraintClass]:
        # https://developer.salesforce.com/docs/atlas.en-us.object_reference.meta/object_reference/field_types.htm#i1435823
        foreignDatasets = [
            builder.make_dataset_urn_with_platform_instance(
                self.platform,
                fsObject,
                self.config.platform_instance,
                self.config.env,
            )
            for fsObject in field["ReferenceTo"]["referenceTo"]
        ]

        for foreignDataset in foreignDatasets:
            yield ForeignKeyConstraintClass(
                name=field["RelationshipName"] if field.get("RelationshipName") else "",
                foreignDataset=foreignDataset,
                foreignFields=[builder.make_schema_field_urn(foreignDataset, "Id")],
                sourceFields=[builder.make_schema_field_urn(datasetUrn, fieldName)],
            )

    def get_report(self) -> SourceReport:
        return self.report


def get_tags(params: Optional[List[str]] = None) -> GlobalTagsClass:
    if params is None:
        params = []
    tags = [TagAssociationClass(tag=builder.make_tag_urn(tag)) for tag in params if tag]
    return GlobalTagsClass(tags=tags)



from __future__ import annotations

import concurrent.futures
from concurrent.futures import Future, ThreadPoolExecutor
from typing import Any, Callable, Iterable, Iterator, Optional, Set, Tuple, TypeVar

_R = TypeVar("_R")


class BackpressureAwareExecutor:
    # This couldn't be a real executor because the semantics of submit wouldn't really make sense.
    # In this variant, if we blocked on submit, then we would also be blocking the thread that
    # we expect to be consuming the results. As such, I made it accept the full list of args
    # up front, and that way the consumer can read results at its own pace.

    @classmethod
    def map(
        cls,
        fn: Callable[..., _R],
        args_list: Iterable[Tuple[Any, ...]],
        max_workers: int,
        max_pending: Optional[int] = None,
    ) -> Iterator[Future[_R]]:
        """Similar to concurrent.futures.ThreadPoolExecutor#map, except that it won't run ahead of the consumer.

        The main benefit is that the ThreadPoolExecutor isn't stuck holding a ton of result
        objects in memory if the consumer is slow. Instead, the consumer can read the results
        at its own pace and the executor threads will idle if they need to.

        Args:
            fn: The function to apply to each input.
            args_list: The list of inputs. In contrast to the builtin map, this is a list
                of tuples, where each tuple is the arguments to fn.
            max_workers: The maximum number of threads to use.
            max_pending: The maximum number of pending results to keep in memory.
                If not set, it will be set to 2*max_workers.

        Returns:
            An iterable of futures.

            This differs from a traditional map because it returns futures
            instead of the actual results, so that the caller is required
            to handle exceptions.

            Additionally, it does not maintain the order of the arguments.
            If you want to know which result corresponds to which input,
            the mapped function should return some form of an identifier.
        """

        if max_pending is None:
            max_pending = 2 * max_workers
        assert max_pending >= max_workers

        pending_futures: Set[Future] = set()

        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            for args in args_list:
                # If the pending list is full, wait until one is done.
                if len(pending_futures) >= max_pending:
                    (done, _) = concurrent.futures.wait(
                        pending_futures, return_when=concurrent.futures.FIRST_COMPLETED
                    )
                    for future in done:
                        pending_futures.remove(future)

                        # We don't want to call result() here because we want the caller
                        # to handle exceptions/cancellation.
                        yield future

                # Now that there's space in the pending list, enqueue the next task.
                pending_futures.add(executor.submit(fn, *args))

            # Wait for all the remaining tasks to complete.
            for future in concurrent.futures.as_completed(pending_futures):
                pending_futures.remove(future)
                yield future

            assert not pending_futures
